import unittest
import io
from tokenizer.jsontokenizer import JsonTokenizer


class TestJsonTokenizerMethods(unittest.TestCase):

    def test_next(self):
        json = '{"name1": "value1", "name2": "value2"}'

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('name1', 1))
        self.assertEqual(tokenizer.next(), ('value1', 1))
        self.assertEqual(tokenizer.next(), ('name2', 1))
        self.assertEqual(tokenizer.next(), ('value2', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.has_next(), False)

    def test_next_with_different_value_types(self):
        json = '{"name1": "value1", "name2": null, "name3": true, "name4": false, "name5": 10.67}'

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('name1', 1))
        self.assertEqual(tokenizer.next(), ('value1', 1))
        self.assertEqual(tokenizer.next(), ('name2', 1))
        self.assertEqual(tokenizer.next(), ('null', 2))
        self.assertEqual(tokenizer.next(), ('name3', 1))
        self.assertEqual(tokenizer.next(), ('true', 2))
        self.assertEqual(tokenizer.next(), ('name4', 1))
        self.assertEqual(tokenizer.next(), ('false', 2))
        self.assertEqual(tokenizer.next(), ('name5', 1))
        self.assertEqual(tokenizer.next(), ('10.67', 2))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.has_next(), False)

    def test_next_simple_array(self):
        json = '["value1", "value2", "value3"]'

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('value1', 1))
        self.assertEqual(tokenizer.next(), ('value2', 1))
        self.assertEqual(tokenizer.next(), ('value3', 1))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.has_next(), False)

    def test_next_nested(self):
        json = '{"name1": "value1", "name2": {"name22": "value22", "name23": "value23"}}'

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('name1', 1))
        self.assertEqual(tokenizer.next(), ('value1', 1))
        self.assertEqual(tokenizer.next(), ('name2', 1))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('name22', 1))
        self.assertEqual(tokenizer.next(), ('value22', 1))
        self.assertEqual(tokenizer.next(), ('name23', 1))
        self.assertEqual(tokenizer.next(), ('value23', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.has_next(), False)

    def test_next_nested_array(self):
        json = '{"name1": "value1", "name2": [10, 20, true, "some value"]}'

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('name1', 1))
        self.assertEqual(tokenizer.next(), ('value1', 1))
        self.assertEqual(tokenizer.next(), ('name2', 1))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('10', 2))
        self.assertEqual(tokenizer.next(), ('20', 2))
        self.assertEqual(tokenizer.next(), ('true', 2))
        self.assertEqual(tokenizer.next(), ('some value', 1))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.has_next(), False)

    def test_next_nested_mixedarray(self):
        json = """
           {
               "name1": "value1",
              "name2": [
                   10,
                   20,
                   {
                       "name3": "value3",
                       "name4": "value4"
                   },
                   30]
           }"""

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('name1', 1))
        self.assertEqual(tokenizer.next(), ('value1', 1))
        self.assertEqual(tokenizer.next(), ('name2', 1))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('10', 2))
        self.assertEqual(tokenizer.next(), ('20', 2))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('name3', 1))
        self.assertEqual(tokenizer.next(), ('value3', 1))
        self.assertEqual(tokenizer.next(), ('name4', 1))
        self.assertEqual(tokenizer.next(), ('value4', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('30', 2))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.has_next(), False)

    def test_next_deeply_nested(self):
        json = """
       {
           "id": "0001",
           "type": "donut",
           "name": "Cake",
           "ppu": 0.55,
           "batters":
           {
               "batter":
                   [
                       { "id": "1001", "type": "Regular" },
                       { "id": "1002", "type": "Chocolate" },
                       { "id": "1003", "type": "Blueberry" },
                       { "id": "1004", "type": "Devils Food" }
                   ]
           },
           "topping":
               [
                   { "id": "5001", "type": "None" },
                   { "id": "5002", "type": "Glazed" },
                   { "id": "5005", "type": "Sugar" },
                   { "id": "5007", "type": "Powdered Sugar" },
                   { "id": "5006", "type": "Chocolate with Sprinkles" },
                   { "id": "5003", "type": "Chocolate" },
                   { "id": "5004", "type": "Maple" }
               ]
       }"""

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('0001', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('donut', 1))
        self.assertEqual(tokenizer.next(), ('name', 1))
        self.assertEqual(tokenizer.next(), ('Cake', 1))
        self.assertEqual(tokenizer.next(), ('ppu', 1))
        self.assertEqual(tokenizer.next(), ('0.55', 2))
        self.assertEqual(tokenizer.next(), ('batters', 1))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('batter', 1))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('1001', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Regular', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('1002', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Chocolate', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('1003', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Blueberry', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('1004', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Devils Food', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('topping', 1))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('5001', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('None', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('5002', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Glazed', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('5005', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Sugar', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('5007', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Powdered Sugar', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('5006', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Chocolate with Sprinkles', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('5003', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Chocolate', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('5004', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('Maple', 1))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.has_next(), False)

    def test_next_array_of_arrays(self):
        json = """
       {
           "id": "123",
           "type": "car",
           "data":
           [
               ["red", "blue", "green"],
               [1,2,3],
               [true, false, true]   
           ]
       }"""

        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)

        self.assertEqual(tokenizer.next(), ('{', 0))
        self.assertEqual(tokenizer.next(), ('id', 1))
        self.assertEqual(tokenizer.next(), ('123', 1))
        self.assertEqual(tokenizer.next(), ('type', 1))
        self.assertEqual(tokenizer.next(), ('car', 1))
        self.assertEqual(tokenizer.next(), ('data', 1))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('red', 1))
        self.assertEqual(tokenizer.next(), ('blue', 1))
        self.assertEqual(tokenizer.next(), ('green', 1))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('1', 2))
        self.assertEqual(tokenizer.next(), ('2', 2))
        self.assertEqual(tokenizer.next(), ('3', 2))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), ('[', 0))
        self.assertEqual(tokenizer.next(), ('true', 2))
        self.assertEqual(tokenizer.next(), ('false', 2))
        self.assertEqual(tokenizer.next(), ('true', 2))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), (']', 0))
        self.assertEqual(tokenizer.next(), ('}', 0))
        self.assertEqual(tokenizer.has_next(), False)


suite = unittest.TestLoader().loadTestsFromTestCase(TestJsonTokenizerMethods)
unittest.TextTestRunner(verbosity=2).run(suite)



from tokenizer.jsontokenizer import JsonTokenizer
from typing import Dict
import stream_json_parser_utils


class JsonDecoder:
    _tokenizer: JsonTokenizer = None
    _class_name: str = 'Root'
    _handler = lambda e, p: e
    _predicate: str = None
    _translators: Dict = None
    _snake_case: bool = False

    def __init__(self):
        self._translators = dict()
        self._snake_case = False

    def tokenizer(self, tokenizer):
        self._tokenizer = tokenizer
        return self

    def root_class_name(self, class_name):
        self._class_name = class_name
        return self

    def event_handler(self, handler):
        self._handler = handler
        return self

    def predicate(self, predicate):
        self._predicate = predicate
        return self

    def translator(self, path, translator):
        self._translators[path] = translator
        return self

    def with_snake_cased_props(self):
        self._snake_case = True
        return self

    def decode(self):
        path = []
        token, token_type = self._tokenizer.next()
        if token == '{':
            self._decode(self._class_name, path,
                         self._predicate.split('.') if self._predicate and len(self._predicate) > 0 else list(),
                         self._handler)
        elif token == '[':
            self._decode_list(self._class_name, path,
                              self._predicate.split('.') if self._predicate and len(self._predicate) > 0 else list(),
                              self._handler)

    def _decode(self, class_name, path, predicate, handler):
        props = dict()
        while self._tokenizer.has_next():
            (prop_name, prop_type) = self._tokenizer.next()
            if prop_name == '}':
                if len(path) >= len(predicate):
                    obj = type(class_name, (), props)()
                    p = '.'.join(path)
                    if p in self._translators:
                        # obj = self._translators[p].translate(obj)
                        obj = self._translators[p](obj)
                    handler(obj, p)
                return
            path.append(prop_name)
            # print(path)
            (prop_value, value_type) = self._tokenizer.next()
            if value_type == 2:
                if prop_value == 'null':
                    prop_value = None
                elif prop_value == 'true':
                    prop_value = True
                elif prop_value == 'false':
                    prop_value = False
            if prop_value == '{':
                if self._path_matches(path, predicate):
                    self._decode(self.to_pascal_case(prop_name), path, predicate, handler)
                elif len(path) > len(predicate):
                    elements = list()
                    self._decode(self.to_pascal_case(prop_name), path, predicate,
                                 lambda e, p: self._record_result(elements, e))
                    prop_value = elements.pop()
                else:
                    self._skip_value()
            elif prop_value == '[':
                if self._path_matches(path, predicate):
                    self._decode_list(self.to_pascal_case(prop_name), path, predicate, handler)
                elif len(path) > len(predicate):
                    elements = list()
                    self._decode_list(self.to_pascal_case(prop_name), path, predicate,
                                      lambda e, p: self._record_result(elements, e))
                    prop_value = elements
                else:
                    self._skip_list()
            else:
                if self._path_matches(path, predicate) and len(path) == len(predicate):
                    p = '.'.join(path)
                    if p in self._translators:
                        # prop_value = self._translators[p].translate(prop_value)
                        prop_value = self._translators[p](prop_value)
                    handler(prop_value, p)

            if len(path) > len(predicate):
                props[self.to_snake_case(prop_name) if self._snake_case else prop_name] = prop_value
            path.pop()

    def _decode_list(self, class_name, path, predicate, handler):
        p = None
        while self._tokenizer.has_next():
            elem, elem_type = self._tokenizer.next()
            if elem == ']':
                break
            if elem == '{':
                self._decode(class_name, path, predicate, handler)
            elif elem == '[':
                if self._path_matches(path, predicate):
                    self._decode_list(class_name, path, predicate, handler)
                elif len(path) > len(predicate):
                    elements = list()
                    self._decode_list(class_name, path, predicate, lambda e, p: self._record_result(elements, e))
                    handler(elements, p)
                else:
                    self._skip_list()
            else:
                if elem_type == 2:
                    if elem == 'null':
                        elem = None
                    elif elem == 'true':
                        elem = True
                    elif elem == 'false':
                        elem = False
                if not p:
                    p = '.'.join(path)
                if p in self._translators:
                    # elem = self._translators[p].translate(elem)
                    elem = self._translators[p](elem)
                handler(elem, p)

    def _skip_value(self):
        while self._tokenizer.has_next():
            prop_name, prop_type = self._tokenizer.next()
            if prop_name == '}':
                return
            prop_value, prop_type = self._tokenizer.next()
            if prop_value == '{':
                self._skip_value()
            elif prop_value == '[':
                self._skip_list()

    def _skip_list(self):
        while self._tokenizer.has_next():
            elem, elem_type = self._tokenizer.next()
            if elem == ']':
                break
            if elem == '{':
                self._skip_value()

    def _record_result(self, elements, e):
        elements.append(e)

    def to_snake_case(self, s):
        return stream_json_parser_utils.to_snake_case(s)

    def to_pascal_case(self, s):
        return stream_json_parser_utils.to_pascal_case(s)

    def _path_matches(self, path, predicate):
        return True if len(predicate) > 0 and len(predicate) >= len(path) and path[-1] == predicate[
            len(path) - 1] else False



import unittest
import io
from tokenizer.jsontokenizer import JsonTokenizer
from decoder.streamdecoder import JsonDecoder


class TestJsonDecoderMethods(unittest.TestCase):

    def test_decode(self):
        json = '{"name1": "value1", "name2": "value2"}'

        result = list()
        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)
        JsonDecoder() \
            .tokenizer(tokenizer) \
            .root_class_name('Example') \
            .event_handler(lambda e, p: result.append((e, p))) \
            .decode()

        self.assertEqual(len(result), 1)

        (obj, path) = result[0]
        self.assertEqual(path, '')
        self.assertEqual(type(obj).__name__, 'Example')
        self.assertEqual(obj.name1, 'value1')
        self.assertEqual(obj.name2, 'value2')

    def test_decode_subscribe_on_nested_object(self):
        json = """
       [
           {
               "id": "0001",
               "type": "donut",
               "name": "Cake",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" },
                               { "id": "1003", "type": "Blueberry" },
                               { "id": "1004", "type": "Devil's Food" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5007", "type": "Powdered Sugar" },
                       { "id": "5006", "type": "Chocolate with Sprinkles" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0002",
               "type": "donut",
               "name": "Raised",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0003",
               "type": "donut",
               "name": "Old Fashioned",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           }
       ]"""

        result = list()
        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)
        JsonDecoder() \
            .tokenizer(tokenizer) \
            .root_class_name('Example') \
            .predicate('batters') \
            .event_handler(lambda e, p: result.append((e, p))) \
            .decode()

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result[0][0].batter), 4)
        self.assertEqual(len(result[1][0].batter), 1)
        self.assertEqual(len(result[2][0].batter), 2)

        (obj, path) = result[0]
        self.assertEqual(path, 'batters')

        self.assertEqual(type(obj).__name__, 'Batters')
        self.assertEqual(type(obj.batter).__name__, 'list')
        self.assertEqual(len(obj.batter), 4)
        self.assertEqual(type(obj.batter[0]).__name__, 'Batter')

        self.assertEqual(obj.batter[0].id, '1001')
        self.assertEqual(obj.batter[0].type, 'Regular')

        self.assertEqual(obj.batter[1].id, '1002')
        self.assertEqual(obj.batter[1].type, 'Chocolate')

        self.assertEqual(obj.batter[2].id, '1003')
        self.assertEqual(obj.batter[2].type, 'Blueberry')

        self.assertEqual(obj.batter[3].id, '1004')
        self.assertEqual(obj.batter[3].type, "Devil's Food")

    def test_decode_subscribe_on_nested_array(self):
        json = """
       [
           {
               "id": "0001",
               "type": "donut",
               "name": "Cake",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" },
                               { "id": "1003", "type": "Blueberry" },
                               { "id": "1004", "type": "Devil's Food" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5007", "type": "Powdered Sugar" },
                       { "id": "5006", "type": "Chocolate with Sprinkles" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0002",
               "type": "donut",
               "name": "Raised",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0003",
               "type": "donut",
               "name": "Old Fashioned",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           }
       ]"""

        result = list()
        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)
        JsonDecoder() \
            .tokenizer(tokenizer) \
            .root_class_name('Example') \
            .predicate('batters.batter') \
            .event_handler(lambda e, p: result.append((e, p))) \
            .decode()

        self.assertEqual(len(result), 7)

        self.assertEqual(result[1][1], 'batters.batter')
        self.assertEqual(result[6][1], 'batters.batter')

        self.assertEqual(type(result).__name__, 'list')
        (obj1, path1) = result[0]
        self.assertEqual(path1, 'batters.batter')
        self.assertEqual(type(obj1).__name__, 'Batter')
        self.assertEqual(obj1.id, '1001')
        self.assertEqual(obj1.type, 'Regular')

        (obj6, path6) = result[6]
        self.assertEqual(path6, 'batters.batter')
        self.assertEqual(type(obj6).__name__, 'Batter')
        self.assertEqual(obj6.id, '1002')
        self.assertEqual(obj6.type, 'Chocolate')

    def test_decode_subscribe_on_simple_value(self):
        json = """
       [
           {
               "id": "0001",
               "type": "donut",
               "name": "Cake",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" },
                               { "id": "1003", "type": "Blueberry" },
                               { "id": "1004", "type": "Devil's Food" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5007", "type": "Powdered Sugar" },
                       { "id": "5006", "type": "Chocolate with Sprinkles" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0002",
               "type": "donut",
               "name": "Raised",
               "ppu": 0.56,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0003",
               "type": "donut",
               "name": "Old Fashioned",
               "ppu": 0.57,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           }
       ]"""

        result = list()
        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)
        JsonDecoder() \
            .tokenizer(tokenizer) \
            .root_class_name('Example') \
            .predicate('ppu') \
            .event_handler(lambda e, p: result.append((e, p))) \
            .decode()

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0][0], '0.55')
        self.assertEqual(result[1][0], '0.56')
        self.assertEqual(result[2][0], '0.57')

    def test_decode_with_translators(self):
        class MyBatter:
            __slots__ = ('my_id', 'my_type')

            def __init__(self, id, type):
                self.my_id = id
                self.my_type = type

        class MyBatters:
            __slots__ = ('my_batter')

            def __init__(self, batter):
                self.my_batter = batter

        json = """
       [
           {
               "id": "0001",
               "type": "donut",
               "name": "Cake",
               "ppu": 0.55,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" },
                               { "id": "1003", "type": "Blueberry" },
                               { "id": "1004", "type": "Devil's Food" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5007", "type": "Powdered Sugar" },
                       { "id": "5006", "type": "Chocolate with Sprinkles" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0002",
               "type": "donut",
               "name": "Raised",
               "ppu": 0.56,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5005", "type": "Sugar" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           },
           {
               "id": "0003",
               "type": "donut",
               "name": "Old Fashioned",
               "ppu": 0.57,
               "batters":
                   {
                       "batter":
                           [
                               { "id": "1001", "type": "Regular" },
                               { "id": "1002", "type": "Chocolate" }
                           ]
                   },
               "topping":
                   [
                       { "id": "5001", "type": "None" },
                       { "id": "5002", "type": "Glazed" },
                       { "id": "5003", "type": "Chocolate" },
                       { "id": "5004", "type": "Maple" }
                   ]
           }
       ]"""

        result = list()
        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)
        JsonDecoder() \
            .tokenizer(tokenizer) \
            .root_class_name('Example') \
            .translator('batters', lambda o: MyBatters(o.batter)) \
            .translator('batters.batter', lambda o: MyBatter(o.id, o.type)) \
            .event_handler(lambda e, p: result.append(e)) \
            .decode()

        self.assertEqual(len(result), 3)
        self.assertEqual(type(result[0].batters).__name__, 'MyBatters')
        self.assertEqual(type(result[0].batters.my_batter[0]).__name__, 'MyBatter')
        self.assertEqual(result[0].batters.my_batter[0].my_id, '1001')
        self.assertEqual(result[0].batters.my_batter[0].my_type, 'Regular')

    def test_decode_array_of_arrays(self):
        json = """
               {
                   "id": "123",
                   "type": "car",
                   "data":
                   [
                       ["red", "blue", "green"],
                       [1,2,3],
                       [true, false, true]   
                   ]
               }"""

        result = list()
        stream = io.StringIO(json)
        tokenizer = JsonTokenizer(stream)
        JsonDecoder() \
            .tokenizer(tokenizer) \
            .root_class_name('Example') \
            .event_handler(lambda e, p: result.append(e)) \
            .decode()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].id, '123')
        self.assertEqual(result[0].type, 'car')
        self.assertEqual(len(result[0].data), 3)
        self.assertEqual(len(result[0].data[0]), 3)
        self.assertEqual(result[0].data[0][0], 'red')
        self.assertEqual(result[0].data[0][1], 'blue')
        self.assertEqual(result[0].data[0][2], 'green')
        self.assertEqual(len(result[0].data[1]), 3)
        self.assertEqual(result[0].data[1][0], '1')
        self.assertEqual(result[0].data[1][1], '2')
        self.assertEqual(result[0].data[1][2], '3')
        self.assertEqual(len(result[0].data[2]), 3)
        self.assertEqual(result[0].data[2][0], True)
        self.assertEqual(result[0].data[2][1], False)
        self.assertEqual(result[0].data[2][2], True)


suite = unittest.TestLoader().loadTestsFromTestCase(TestJsonDecoderMethods)
unittest.TextTestRunner(verbosity=2).run(suite)



from setuptools import setup, Extension
import setuptools

stream_json_parser_utils = Extension('stream_json_parser_utils', extra_compile_args = [ '--std=c++11' ], sources = [ 'utils/src/utils.cpp' ])

if __name__ == '__main__':
   setup(
       name='streamjsonparser',
       version='1.0',
       description='Stream Json Parser',
       ext_modules=[stream_json_parser_utils],
       author='Andrey Lysenko',
       author_email='andrey.lysenko@gmail.com',
       packages=setuptools.find_packages(),
       include_package_data=True,
       zip_safe=False
   )



import io
import ctypes
import stream_json_parser_utils


class JsonTokenizer:
    _stream: io.RawIOBase = None
    _encoding = None
    _buffer_size: int = None
    _buffer = None
    _buffer_pos = ctypes.c_int(-1)
    _simple_value_started = ctypes.c_bool(False)
    _need_to_decode = None

    def __init__(self, stream: io.RawIOBase, encoding: str = 'utf-8', buffer_size: int = 20):
        self._stream = stream
        if isinstance(self._stream, io.TextIOBase):
            self._need_to_decode = False
        else:
            self._need_to_decode = True
        self._encoding = encoding
        self._buffer_size = buffer_size
        self._buffer = ''
        self._buffer_pos = ctypes.c_int(-1)
        self._simple_value_started = ctypes.c_bool(False)

    def next(self) -> str:
        (buffer, self._buffer_size, self._buffer_pos.value, token, token_type,
         self._simple_value_started.value) = stream_json_parser_utils.next(self._stream, self._buffer_size,
                                                                           self._need_to_decode, self._encoding,
                                                                           self._buffer, self._buffer_pos.value,
                                                                           self._simple_value_started.value)
        if buffer:
            self._buffer = buffer
        return token, token_type

    def has_next(self) -> bool:
        (buffer, self._buffer_size, self._buffer_pos.value, hasnext) = stream_json_parser_utils.has_next(self._stream,
                                                                                                         self._buffer_size,
                                                                                                         self._need_to_decode,
                                                                                                         self._encoding,
                                                                                                         self._buffer,
                                                                                                         self._buffer_pos.value)
        if buffer:
            self._buffer = buffer
        return hasnext
